{-# OPTIONS_GHC -Wno-unrecognised-pragmas #-}
{-# HLINT ignore "Avoid lambda" #-}
{-# HLINT ignore "Eta reduce" #-}
module Sudoku where
import Data.Char (digitToInt)
import Data.List (transpose, elemIndex)
import Data.Set (Set, fromList, toList, member, size, difference, unions, lookupGT, deleteAt)
import Data.Maybe (fromJust)
import Control.Applicative ((<|>))

splitList :: Int -> [a] -> [[a]]
splitList _ [] = []
splitList n oriList = prev : splitList n next
  where
    (prev, next) = splitAt n oriList

example :: String
example = "000000021430000000600000000201500000000006370000000000068000400000230000000070000"

{-
lineToList :: [Char] -> [Int]
Given a string which contains 81 numbers (the original sudoku),
return the list form of it.
-}
lineToList :: [Char] -> [Int]
lineToList oriLine = map digitToInt oriLine

{-
getCell :: Int -> [a] -> [a]
Given a Cell number and a grid ordered by row, return all 9 elements in this cell. 
We use the formula (n/3)*9 + n%3 to get the first row of the cell, 
add the number 3 to get the second row of the cell, 
and add the number 6 to get the third row of the cell.
-}
getCell :: Int -> [a] -> [a]
getCell n oriList = newList !! cellIndex ++ newList !! (cellIndex + 3) ++ newList !! (cellIndex + 6)
  where
    cellIndex = div n 3 * 9 + mod n 3
    newList = splitList 3 oriList

{-
getRow :: Int -> [a] -> [a]
Given a row number and a grid ordered by cell, return all 9 elements in this row. 
We use the formula n%3 + (n/3)*9 to get the first three elements in the nth row, 
add the number 3 to get the second three elements in the nth row, 
and add the number 6 to get the last three elements in the nth row.
-}
getRow :: Int -> [a] -> [a]
getRow n cellGrid = newList !! rowIndex ++ newList !! (rowIndex + 3) ++ newList !! (rowIndex + 6)
  where
    rowIndex = mod n 3 + (div n 3) * 9
    newList = splitList 3 cellGrid


{-
getRowGrid :: [a] -> [[a]]
Given the original list of elements, 
return a new list that contains sublists of 9 elements. 
Each sublist represents a row in the sudoku.
-}
getRowGrid :: [a] -> [[a]]
getRowGrid oriList = splitList 9 oriList

{-
getColGrid :: [a] -> [[a]]
Given the original list of elements, 
return a new list that contains sublists of 9 elements. 
Each sublist represents a column in the sudoku.
-}
getColGrid :: [a] -> [[a]]
getColGrid oriList = transpose $ getRowGrid oriList

{-
getCellGrid :: [a] -> [[a]]
Given the original list of elements, 
return a new list that contains sublists of 9 elements. 
Each sublist represents a cell in the sudoku.
-}

getCellGrid :: [a] -> [[a]]
getCellGrid oriList = [ getCell i oriList | i <- [0..8] ]

{-
showGrid :: [Int] -> IO ()
Given the orignal 81 elements list, 
display that in the 9x9 sudoku form.
-}
showGrid :: [Int] -> IO ()
showGrid oriList = mapM_ print (getRowGrid oriList)


{-
possibleGrid :: (Ord a, Num a, Enum a) => [a] -> [Set a]
Given the original list, 
return a list of sets that contains all possible values for each unfilled number.
-}
possibleGrid :: (Ord a, Num a, Enum a) => [a] -> [Set a]
possibleGrid oriList = [ if member val def then fromList [val] else def | val <- oriList]
  where
    def = fromList [1..9]


{-
getFixedByRow :: Ord a => [Set a] -> [Set a]
Given the original list of elements, return a list sets.
Each set contains the fixed value (which is given at beginning) in each row.
getFixedByCell :: Ord a => [Set a] -> [Set a]
Given the original list of elements, return a list sets.
Each set contains the fixed value (which is given at beginning) in each cell.
getFixedByCol :: [Int] -> [Set Int]
Given the original list of elements, return a list sets.
Each set contains the fixed value (which is given at beginning) in each column.
-}
getFixedByRow :: Ord a => [Set a] -> [Set a]
getFixedByRow possGrid = [ unions $ filter (\x -> size x == 1) row | row <- getRowGrid possGrid ]

getFixedByCell :: Ord a => [Set a] -> [Set a]
getFixedByCell possGrid = [ unions $ filter (\x -> size x == 1) row | row <- getCellGrid possGrid ]

getFixedByCol :: Ord a => [Set a] -> [Set a]
getFixedByCol possGrid = [ unions $ filter (\x -> size x == 1) row | row <- getColGrid possGrid ]


{-
hardPruneHelper :: Ord a => [[Set a]] -> [Set a] -> [[Set a]]
Given a list of sets of all possible value in rows/columns/cells and a list of sets of fixed numbers in rows/columns/cells,
return all possible value except the fixed value in rows/columns/cells.
hardPruneEach :: Ord a => [Set a] -> [Set a]
Use hardPruneEach to eliminate fixed values for all numbers. 
hardPrune repeats the process until there is no way to further eliminate the possibilities.
Given a list of original list, return a list of sets contains possible value for each number.
-}
hardPruneHelper :: Ord a => [[Set a]] -> [Set a] -> [[Set a]]
hardPruneHelper allSet fixedRowSet = [ map (\x -> if size x/=1 then x `difference` f else x) r | (r,f) <- match ]
  where
    match = zip allSet fixedRowSet

hardPruneEach :: Ord a => [Set a] -> [Set a]
hardPruneEach possGrid = concat [ getRow i (concat thiPrune) | i <- [0..8] ]
  where 
    fstPrune = hardPruneHelper (getRowGrid possGrid) (getFixedByRow possGrid)
    sndPrune = hardPruneHelper (getColGrid (concat fstPrune)) (getFixedByCol possGrid)
    thiPrune = hardPruneHelper (getCellGrid (concat $ transpose sndPrune)) (getFixedByCell possGrid)

hardPrune :: Ord a => [Set a] -> [Set a]
hardPrune possGrid | possGrid ==  hardPruneEach possGrid = possGrid
                   | otherwise = hardPruneEach possGrid


{-
softPrune :: [Set a] -> [[Set a]]
Given a possibility grid (list of 81 sets where each set indicates all possible values for each square),
find the square with least number of possibilities, choose a possible value,
and return a list of newly-generated possiblity grid.
-}
softPrune :: Ord a => [Set a] -> ([Set a], [Set a])
softPrune poss | minSize == Nothing = (poss, poss)
               | otherwise = (chosenGrid, unchosenGrid)
  where
    (prev, mid : next) = splitAt index poss
    sizeGrid = map size poss
    minSize = lookupGT 1 (fromList sizeGrid)
    index = fromJust $ elemIndex (fromJust minSize) sizeGrid
    chosenGrid = prev ++ [fromList [head $ toList mid]] ++ next
    unchosenGrid = prev ++ [deleteAt 0 mid] ++ next

{-
ifSolved :: (Ord a, Num a, Enum a) => [Set a] -> Bool
Given a possiblity grid, returns true if the sudoku is solved (that is, 
all rows, columns, and cells contain exactly nine increasing numbers 
eg: 1,2,3,4,5,6,7,8,9); returns false if the sudoku is not solved
-}
ifSolved :: (Ord a, Num a, Enum a) => [Set a] -> Bool
ifSolved poss = and [unions row == fromList [1..9]| row <- getRowGrid poss]
      && and [unions col == fromList [1..9]| col <- getColGrid poss]
      && and [unions cell == fromList [1..9]| cell <- getCellGrid poss]
      && map (\x -> size x) poss == take 81 [1,1..]


{-
ifValid :: (Ord a, Num a) => [Set a] -> Bool
Given a possiblity grid, 
returns true if all values in the newly-generated grid is consistent,
and returns false if any condition is not met.
-}
ifValid :: (Ord a, Num a) => [Set a] -> Bool
ifValid poss = and [s /= 0 | s <- map (\x -> size x) poss] && and boolList
  where possList = map (\x -> if size x > 1 then -1 else head $ toList x) poss
        allList = getColGrid possList ++ getRowGrid possList ++ getCellGrid possList
        boolList = [ length l == size (fromList l) | list <- allList, let l = filter (/= (-1)) list]


{-
possToGrid :: [Set b] -> [b]
A helper function that converts possible grid back to grid.
solveSudoku :: (Num a, Enum a, Ord a) => [Set a] -> Maybe [a]
Given a possibility grid, implement backtracking search using the minimum remaining value heuristic.
and apply forward checking to reduce variables domains.
If a solution is found, return the list of values in such a grid; else report the error.
solve :: [Char] -> Maybe [Int]
read different forms of input sudoku and call sudoku solver to get the result. 
-}
possToGrid :: [Set b] -> [b]
possToGrid poss = map (\x -> head $ toList x) poss

solveSudoku :: (Num a, Enum a, Ord a) => [Set a] -> Maybe [a]
solveSudoku poss | ifSolved poss = Just (possToGrid poss)
                 | not $ ifValid poss = Nothing
                 | otherwise = solveSudoku (hardPrune chosen) <|> solveSudoku (hardPrune unchosen)
  where
    (chosen, unchosen) = softPrune poss

solve :: [Char] -> Maybe [Int]
solve text = solveSudoku grid
  where
    replace = map (\c -> if c=='.' then '0'; else c)
    editedText = replace text
    grid = possibleGrid $ lineToList editedText
